using System;
using Cairo;
using Vintagestory.API.MathTools;
using Vintagestory.API.Client;
using Vintagestory.API.Common;

namespace TemporalStormTimer {
	// This is based on GuiElementStatbar because the original is broken and unfixable through inheritance.
	public class TemporalStormStatbar: GuiElementTextBase {
		protected const int gearSize = 64;

		public bool shouldFlash;
		public float flashTime;

		protected ICoreClientAPI capi;

		protected double[] colour;
		protected bool rightToLeft = false;

		protected bool visible = false;
		protected float minValue = 0;
		protected float maxValue = 100;
		protected float value = 0;
		protected bool valuesSet;

		protected float lineInterval = 10;

		protected LoadedTexture baseTexture;
		protected LoadedTexture barTexture;
		protected LoadedTexture flashTexture;
		protected LoadedTexture valueTexture;

		protected MeshRef gearMesh;
		protected LoadedTexture gearTexture;
		protected Matrixf gearMatrix;
		protected float gearAngle = 0;
		protected float gearOffsetY = 0;
		protected bool spinGear;

		private int valueHeight;

		public TemporalStormStatbar(ICoreClientAPI capi, ElementBounds bounds, double[] colour, bool rightToLeft, bool spinGear): base(capi, "", CairoFont.WhiteDetailText(), bounds) {
			this.capi = capi;

			this.barTexture = new LoadedTexture(capi);
			this.flashTexture = new LoadedTexture(capi);
			this.valueTexture = new LoadedTexture(capi);
			this.baseTexture = new LoadedTexture(capi);

			this.gearTexture = new LoadedTexture(capi);
			this.gearMesh = capi.Render.UploadMesh(QuadMeshUtil.GetQuad());
			this.gearMatrix = new Matrixf();
			this.gearOffsetY = (float)(bounds.fixedHeight / 2f);

			this.colour = colour;
			this.rightToLeft = rightToLeft;
			this.spinGear = spinGear;
		}

		public override void ComposeElements(Context ctx, ImageSurface surface) {
			Bounds.CalcWorldBounds();

			surface = new ImageSurface(Format.Argb32, Bounds.OuterWidthInt+1, Bounds.OuterHeightInt + 1);
			ctx = new Context(surface);

			RoundRectangle(ctx, 0, 0, Bounds.InnerWidth, Bounds.InnerHeight, 1);

			ctx.SetSourceRGBA(0.15, 0.15, 0.15, 1);
			ctx.Fill();
			EmbossRoundRectangleElement(ctx, 0, 0, Bounds.InnerWidth, Bounds.InnerHeight, false, 3, 1);

			if(this.valuesSet) {
				this.RecomposeOverlays();
			}

			generateTexture(surface, ref this.baseTexture);
			surface.Dispose();
			ctx.Dispose();
		}

		public void setVisible(bool visible) {
			this.visible = visible;
			
			if(!this.visible) {
				this.gearAngle = 0;
			}
		}

		protected void RecomposeOverlays() {
			TyronThreadPool.QueueTask(() => {
						ComposeValueOverlay();
						ComposeFlashOverlay();
						ComposeGear();
					});
		}

		protected void ComposeValueOverlay() {
			Bounds.CalcWorldBounds();

			// Fix width formula.
			double widthRel = (double)((this.value - this.minValue) / (this.maxValue - this.minValue));
			this.valueHeight = (int)Bounds.OuterHeight + 1;
			ImageSurface surface = new ImageSurface(Format.Argb32, Bounds.OuterWidthInt + 1, this.valueHeight);
			Context ctx = new Context(surface);

			if(widthRel > 0.01) {
				double width = Bounds.OuterWidth * widthRel;
				double x = this.rightToLeft
						? Bounds.OuterWidth - width
						: 0;

				RoundRectangle(ctx, x, 0, width, Bounds.OuterHeight, 1);
				ctx.SetSourceRGB(this.colour[0], this.colour[1], this.colour[2]);
				ctx.FillPreserve();

				ctx.SetSourceRGB(this.colour[0] * 0.4, this.colour[1] * 0.4, this.colour[2] * 0.4);
				ctx.LineWidth = scaled(3);
				ctx.StrokePreserve();
				surface.BlurFull(3);

				width = Bounds.InnerWidth * widthRel;
				x = this.rightToLeft
						? Bounds.InnerWidth - width
						: 0;

				EmbossRoundRectangleElement(ctx, x, 0, width, Bounds.InnerHeight, false, 2, 1);
			}

			ctx.SetSourceRGBA(0, 0, 0, 0.5);
			ctx.LineWidth = scaled(2.2);

			int lines = Math.Min(50, (int)((this.maxValue - this.minValue) / this.lineInterval));
			
			for(int i = 1; i < lines; i ++) {
				ctx.NewPath();
				ctx.SetSourceRGBA(0, 0, 0, 0.5);

				double x = (Bounds.InnerWidth * i) / lines;

				ctx.MoveTo(x, 0);
				ctx.LineTo(x, Math.Max(3, Bounds.InnerHeight - 1));
				ctx.ClosePath();
				ctx.Stroke();
			}

			api.Event.EnqueueMainThreadTask(() => {
						generateTexture(surface, ref this.barTexture);

						ctx.Dispose();
						surface.Dispose();
					}, "recompstatbar");
		}

		protected void ComposeFlashOverlay() {
			this.valueHeight = (int)Bounds.OuterHeight + 1;

			ImageSurface surface = new ImageSurface(Format.Argb32, Bounds.OuterWidthInt + 28, Bounds.OuterHeightInt + 28);
			Context ctx = new Context(surface);

			ctx.SetSourceRGBA(0, 0, 0, 0);
			ctx.Paint();

			RoundRectangle(ctx, 12, 12, Bounds.OuterWidthInt + 4, Bounds.OuterHeightInt + 4, 1);
			ctx.SetSourceRGB(this.colour[0], this.colour[1], this.colour[2]);
			ctx.FillPreserve();
			surface.BlurFull(3);
			ctx.Fill();
			surface.BlurFull(2);

			RoundRectangle(ctx, 15, 15, Bounds.OuterWidthInt - 2, Bounds.OuterHeightInt - 2, 1);
			ctx.Operator = Operator.Clear;
			ctx.SetSourceRGBA(0, 0, 0, 0);
			ctx.Fill();

			api.Event.EnqueueMainThreadTask(() => {
						generateTexture(surface, ref this.flashTexture);
						ctx.Dispose();
						surface.Dispose();
					}, "recompstatbar");
		}

		protected void ComposeGear() {
			ImageSurface surface = new ImageSurface(Format.Argb32, gearSize, gearSize);
			Context context = new Context(surface);

			capi.Gui.Icons.DrawVSGear(context, surface, 0, 0, gearSize, gearSize, this.colour);
			api.Event.EnqueueMainThreadTask(() => {
						generateTexture(surface, ref this.gearTexture);
						context.Dispose();
						surface.Dispose();
					}, "recompstatbar");
		}

		public override void RenderInteractiveElements(float deltaTime) {
			if(!this.visible) {
				return;
			}

			double x = Bounds.renderX;
			double y = Bounds.renderY;

			this.renderGear(x, y + this.gearOffsetY);

			api.Render.RenderTexture(this.baseTexture.TextureId, x, y, Bounds.OuterWidthInt + 1, Bounds.OuterHeightInt + 1);

			float alpha = 0;
			if(this.shouldFlash) {
				this.flashTime += 6 * deltaTime;
				alpha = GameMath.Sin(this.flashTime);

				if(alpha < 0) {
					this.shouldFlash = false;
					this.flashTime = 0;
				}

				if(this.flashTime < GameMath.PIHALF) {
					alpha = Math.Min(1, alpha * 3);
				}
			}

			if(alpha > 0) {
				api.Render.RenderTexture(this.flashTexture.TextureId, x - 14, y - 14, Bounds.OuterWidthInt + 28, Bounds.OuterHeightInt + 28, 50, new Vec4f(1.5f, 1, 1, alpha));
			}

			if(this.barTexture.TextureId > 0) {
				api.Render.RenderTexture(this.barTexture.TextureId, x, y, Bounds.OuterWidthInt + 1, this.valueHeight);
			}
		}

		protected void renderGear(double x, double y) {
			if(this.gearMesh.Disposed || this.gearTexture.Disposed) {
				return;
			}

			IShaderProgram prog = api.Render.GetEngineShader(EnumShaderProgram.Gui);
			prog.Uniform("rgbaIn", ColorUtil.WhiteArgbVec);
			prog.Uniform("extraGlow", 0);
			prog.Uniform("applyColor", 0);
			prog.Uniform("noTexture", 0f);
			prog.BindTexture2D("tex2d", this.gearTexture.TextureId, 0);

			this.gearMatrix.Set(api.Render.CurrentModelviewMatrix)
				.Translate(x, y, 50)
				.Scale(this.gearTexture.Width, this.gearTexture.Height, 0)
				.Scale((float)GuiElement.scaled(0.5f), (float)GuiElement.scaled(0.5f), 0)
				.RotateZ(this.gearAngle * GameMath.DEG2RAD);
			if(this.spinGear) {
				this.gearAngle += 2;
			}

			prog.UniformMatrix("projectionMatrix", api.Render.CurrentProjectionMatrix);
			prog.UniformMatrix("modelViewMatrix", this.gearMatrix.Values);

			api.Render.RenderMesh(this.gearMesh);
		}

		public void SetLineInterval(float value) {
			this.lineInterval = value;
		}

		public void SetValue(float value) {
			this.value = value;
			this.valuesSet = true;
			this.RecomposeOverlays();
		}

		public float GetValue() {
			return this.value;
		}

		public void SetValues(float value, float min, float max) {
			this.value = value;
			this.minValue = min;
			this.maxValue = max;
			this.valuesSet = true;
			this.RecomposeOverlays();
		}

		public void SetMinMax(float min, float max) {
			this.minValue = min;
			this.maxValue = max;
			this.RecomposeOverlays();
		}
		
		public override void Dispose() {
			base.Dispose();

			this.baseTexture.Dispose();
			this.barTexture.Dispose();
			this.flashTexture.Dispose();
			this.valueTexture.Dispose();

			this.gearMesh.Dispose();
			this.gearTexture.Dispose();
		}
	}
}