using System.Reflection;
using HarmonyLib;
using Vintagestory.API.Client;
using Vintagestory.API.Common;
using Vintagestory.GameContent;

namespace TemporalStormTimer {
	public class TemporalStormTimerSystem: ModSystem {
		public const string harmonyId = "temporalstormtimer";
		public const int listenInterval = 1000;

		protected const double approachingThreshold = 0.35;
		protected const double imminentThreshold = 0.02;
		protected const double waningThreshold = 0.02;

		public static double[] warningColour;
		public static double[] activeColour;
		
		protected TemporalStormTimerConfigManager config;

		protected static ICoreClientAPI capi;
		protected TemporalStormTimerElement element;

		protected long listenerId;
		protected Harmony harmony;

		protected static TemporalStormRunTimeData data;
		protected bool stormActive;
		protected double stormLengthHours;

		public override bool ShouldLoad(EnumAppSide forSide) {
			return forSide == EnumAppSide.Client;
		}

		public override void StartClientSide(ICoreClientAPI capi) {
			base.StartClientSide(capi);
			TemporalStormTimerSystem.capi = capi;

			this.config = new TemporalStormTimerConfigManager(TemporalStormTimerSystem.capi);

			TemporalStormTimerSystem.warningColour = this.convert(this.config.Get().warning_colour);
			TemporalStormTimerSystem.activeColour = this.convert(this.config.Get().active_colour);
			
			this.element = new TemporalStormTimerElement(TemporalStormTimerSystem.capi);
			this.stormActive = false;

			if(TemporalStormTimerSystem.capi.ModLoader.GetModSystem<SystemTemporalStability>() != null) {
				this.listenerId = TemporalStormTimerSystem.capi.Event.RegisterGameTickListener(this.tick, listenInterval);
				this.harmony = new Harmony(harmonyId);

				this.harmony.Patch(typeof(SystemTemporalStability).GetMethod("onServerData", BindingFlags.Instance | BindingFlags.NonPublic),
						postfix: new HarmonyMethod(typeof(TemporalStormTimerSystem).GetMethod(nameof(TemporalStormTimerSystem.clientReceiveData))));
			}
		}

		public override void Dispose() {
			base.Dispose();

			this.harmony.UnpatchAll(harmonyId);
		}

		protected void tick(float dt) {
			if(TemporalStormTimerSystem.data == null) {
				return;
			}

			double nextStormDaysLeft = TemporalStormTimerSystem.data.nextStormTotalDays - TemporalStormTimerSystem.capi.World.Calendar.TotalDays;

			if(nextStormDaysLeft > 0 && nextStormDaysLeft < approachingThreshold) {
				// Preparing.
				this.tickPreparing();
			} else {
				// In progress.
				if(TemporalStormTimerSystem.data.nowStormActive) {
					// Active.
					this.tickActive();
				} else if(this.stormActive) {
					// Ending.
					if(this.config.Get().show_warning) {
						this.element.warningBar.setVisible(false);
					}

					if(this.config.Get().show_active) {
						this.element.activeBar.setVisible(false);
					}

					this.stormActive = false;
				}
			}
		}

		protected void tickPreparing() {
			if(this.config.Get().show_warning) {
				float hoursLeft = (float)((TemporalStormTimerSystem.data.nextStormTotalDays - TemporalStormTimerSystem.capi.World.Calendar.TotalDays) * TemporalStormTimerSystem.capi.World.Calendar.HoursPerDay);
				float approachingHours = (float)(approachingThreshold * TemporalStormTimerSystem.capi.World.Calendar.HoursPerDay);

				this.element.warningBar.setVisible(true);
				this.element.warningBar.SetValues(approachingHours - hoursLeft, 0, approachingHours);
				this.element.warningBar.shouldFlash = hoursLeft <= (TemporalStormTimerSystem.capi.World.Calendar.HoursPerDay * imminentThreshold);
			}

			if(this.config.Get().show_active) {
				this.element.activeBar.setVisible(false);
			}
		}

		protected void tickActive() {
			double hoursLeft = (TemporalStormTimerSystem.data.stormActiveTotalDays - TemporalStormTimerSystem.capi.World.Calendar.TotalDays) * TemporalStormTimerSystem.capi.World.Calendar.HoursPerDay;

			if(!this.stormActive) {
				// Initialize storm length.
				this.stormLengthHours = hoursLeft;
				this.stormActive = true;
			}

			if(this.config.Get().show_warning) {
				this.element.warningBar.setVisible(false);
			}

			if(this.config.Get().show_active) {
				this.element.activeBar.setVisible(true);
				this.element.activeBar.SetValues((float)hoursLeft, 0, (float)this.stormLengthHours);
				this.element.activeBar.shouldFlash = hoursLeft <= (TemporalStormTimerSystem.capi.World.Calendar.HoursPerDay * waningThreshold);
			}
		}

		public static void clientReceiveData(TemporalStormRunTimeData data) {
			TemporalStormTimerSystem.data = data;
		}

		private double[] convert(string hex) {
			int r = int.Parse(hex.Substring(1, 2), System.Globalization.NumberStyles.HexNumber);
			int g = int.Parse(hex.Substring(3, 2), System.Globalization.NumberStyles.HexNumber);
			int b = int.Parse(hex.Substring(5, 2), System.Globalization.NumberStyles.HexNumber);
			int a = hex.Length < 8 ? 255 : int.Parse(hex.Substring(7, 2), System.Globalization.NumberStyles.HexNumber);

			return new double[] { r / 255.0, g / 255.0, b / 255.0, a / 255.0 };
		}
	}
}